# -*- coding: utf-8 -*-
#
# Copyright (c) Panu Lahtinen
#
# Author(s):
#
#   Panu Lahtinen <pnuu+git@iki.fi>
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

import defusedxml.ElementTree as ET
import datetime as dt

import numpy as np

from fmiopendata import wfs
from fmiopendata.utils import read_url

TIME_FORMAT = "%Y-%m-%dT%H:%M:%SZ"
FIELD_NAMES = {"PAP_PT1S_AVG": "pressures",
               "WSP_PT1S_AVG": "wind_speeds",
               "WDP_PT1S_AVG": "wind_directions",
               "TAP_PT1S_AVG": "temperatures",
               "TDP_PT1S_AVG": "dew_points",
               }


class Sounding(object):
    """Class for holding a single sounding."""

    def __init__(self):
        """Initialize sounding."""
        self.name = None
        self.id = None
        self.nominal_time = None
        self.start_time = None
        self.end_time = None
        self.lats = None
        self.lons = None
        self.altitudes = None
        self.times = None
        self.pressures = None
        self.temperatures = None
        self.dew_points = None
        self.wind_speeds = None
        self.wind_directions = None


class ParseSoundings(object):
    """Collect sounding data."""

    def __init__(self, xml):
        """Initialize sounding collector."""
        self._xml = ET.fromstring(xml)
        self.soundings = []
        self._parse()

    def _parse(self):
        """Parse sounding data."""
        for member in self._xml.findall(wfs.WFS_MEMBER):
            sounding = Sounding()
            for name in member.findall(wfs.GML_NAME):
                try:
                    if name.attrib["codeSpace"] == "http://xml.fmi.fi/namespace/locationcode/wmo":
                        sounding.id = name.text
                    elif name.attrib["codeSpace"] == "http://xml.fmi.fi/namespace/locationcode/name":
                        sounding.name = name.text
                except KeyError:
                    continue

            sounding.nominal_time = dt.datetime.strptime(member.findtext(wfs.GML_TIME_POSITION), TIME_FORMAT)
            sounding.start_time = dt.datetime.strptime(member.findtext(wfs.GML_BEGIN_POSITION), TIME_FORMAT)
            sounding.end_time = dt.datetime.strptime(member.findtext(wfs.GML_END_POSITION), TIME_FORMAT)

            try:
                positions = np.fromstring(member.findtext(wfs.GMLCOV_POSITIONS), dtype=float, sep=" ")
            except TypeError:
                print("No soundings found")
                return
            sounding.lats = positions[::4]
            sounding.lons = positions[1::4]
            sounding.altitudes = positions[2::4]
            times = positions[3::4]
            sounding.times = np.array([dt.datetime.utcfromtimestamp(t) for t in times])

            data = np.fromstring(member.findtext(wfs.GML_DOUBLE_OR_NIL_REASON_TUPLE_LIST), dtype=float, sep=" ")
            fields = member.findall(wfs.SWE_FIELD)
            for i, field in enumerate(fields):
                setattr(sounding, FIELD_NAMES[field.attrib["name"]], data[i::len(fields)])

            self.soundings.append(sounding)


def download_and_parse(query_id, args=None):
    """Download and parse the given stored query."""
    url = wfs.STORED_QUERY_URL + query_id
    if args:
        url = url + "&" + "&".join(args)
    xml = read_url(url)
    return ParseSoundings(xml)



# -*- coding: utf-8 -*-
#
# Copyright (c) Panu Lahtinen
#
# Author(s):
#
#   Panu Lahtinen <pnuu+git@iki.fi>
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

import defusedxml.ElementTree as ET

from fmiopendata.utils import read_url


BASE_URL = "https://opendata.fmi.fi/wfs?service=WFS&request="
STORED_QUERY_URL = "https://opendata.fmi.fi/wfs?service=WFS&version=2.0.0&request=getFeature&storedquery_id="

GML_BEGIN_POSITION = ".//{http://www.opengis.net/gml/3.2}beginPosition"
GML_DOUBLE_OR_NIL_REASON_TUPLE_LIST = ".//{http://www.opengis.net/gml/3.2}doubleOrNilReasonTupleList"
GML_END_POSITION = ".//{http://www.opengis.net/gml/3.2}endPosition"
GML_FILE_REFERENCE = ".//{http://www.opengis.net/gml/3.2}fileReference"
GML_ID = "{http://www.opengis.net/gml/3.2}id"
GML_IDENTIFIER = ".//{http://www.opengis.net/gml/3.2}identifier"
GML_MEASURE = ".//{http://www.opengis.net/gml/3.2}Measure"
GML_NAME = ".//{http://www.opengis.net/gml/3.2}name"
GML_POINT = ".//{http://www.opengis.net/gml/3.2}Point"
GML_POS = ".//{http://www.opengis.net/gml/3.2}pos"
GML_TIME_INSTANT = ".//{http://www.opengis.net/gml/3.2}TimeInstant"
GML_TIME_POSITION = ".//{http://www.opengis.net/gml/3.2}timePosition"
GMLCOV_POSITIONS = ".//{http://www.opengis.net/gmlcov/1.0}positions"
LINK = "{http://www.w3.org/1999/xlink}href"
OM_NAME = ".//{http://www.opengis.net/om/2.0}name"
OM_PARAMETER = ".//{http://www.opengis.net/om/2.0}parameter"
OMOP_LABEL = ".//{http://inspire.ec.europa.eu/schemas/omop/2.9}label"
OMOP_UOM = ".//{http://inspire.ec.europa.eu/schemas/omop/2.9}uom"
SWE_DATA_RECORD = ".//{http://www.opengis.net/swe/2.0}DataRecord"
SWE_FIELD = ".//{http://www.opengis.net/swe/2.0}field"
SWE_LABEL = ".//{http://www.opengis.net/swe/2.0}label"
SWE_UOM = ".//{http://www.opengis.net/swe/2.0}uom"
WFS_ABSTRACT = ".//{http://www.opengis.net/wfs/2.0}Abstract"
WFS_MEMBER = ".//{http://www.opengis.net/wfs/2.0}member"
WFS_PARAMETER = ".//{http://www.opengis.net/wfs/2.0}Parameter"
WFS_PARAMETER_NAME = ".//{http://xml.fmi.fi/schema/wfs/2.0}ParameterName"
WFS_PARAMETER_VALUE = ".//{http://xml.fmi.fi/schema/wfs/2.0}ParameterValue"
WFS_RETURN_FEATURE_TYPE = ".//{http://www.opengis.net/wfs/2.0}ReturnFeatureType"
WFS_STORED_QUERY = ".//{http://www.opengis.net/wfs/2.0}StoredQuery"
WFS_TIME = ".//{http://xml.fmi.fi/schema/wfs/2.0}Time"
WFS_TITLE = ".//{http://www.opengis.net/wfs/2.0}Title"


def get_req_xml(req):
    """Get request XML for *req*."""
    return read_url(BASE_URL + req)


def get_capabilities():
    """Get WFS cababilities."""
    xml = get_req_xml("getCapabilities")

    return xml


def get_stored_queries():
    """Get stored queries."""
    queries = get_req_xml("ListStoredQueries")
    res = dict()
    root = ET.fromstring(queries)
    queries = root.findall(WFS_STORED_QUERY)
    for query in queries:
        name = query.attrib['id']
        if "ById" in name:
            continue
        title = query.findtext(WFS_TITLE)
        return_type = query.findtext(WFS_RETURN_FEATURE_TYPE)
        res[name] = dict({'title': title, 'return_type': return_type})

    return res


def get_stored_query_descriptions():
    """Get stored query descriptions."""
    descriptions = get_req_xml("DescribeStoredQueries")
    res = dict()
    root = ET.fromstring(descriptions)
    for f in root:
        if f.attrib['id'] == 'GetDataSetById':
            continue
        f_ch = list(f)
        desc = dict({'title': f_ch[0].text.strip(),
                     'description': f_ch[1].text.strip()})
        res[f.attrib['id']] = desc

    return res


def download_stored_query(query_id, args=None):
    """Download and parse a stored query."""
    if "radar" in query_id.lower():
        from fmiopendata.radar import download_and_parse
    elif "sounding" in query_id.lower():
        from fmiopendata.sounding import download_and_parse
    elif "lightning" in query_id.lower():
        from fmiopendata.lightning import download_and_parse
    elif "grid" in query_id.lower():
        from fmiopendata.grid import download_and_parse
    elif "multipointcoverage" in query_id.lower():
        from fmiopendata.multipoint import download_and_parse
    else:
        raise NotImplementedError("No parser available for %s" % query_id)

    return download_and_parse(query_id, args=args)



# -*- coding: utf-8 -*-
#
# Copyright (c) Panu Lahtinen
#
# Author(s):
#
#   Panu Lahtinen <pnuu+git@iki.fi>
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

from urllib.request import urlretrieve
import warnings

import requests
import defusedxml.ElementTree as ET

EXCEPTION_TEXT = './/{http://www.opengis.net/ows/1.1}ExceptionText'


def read_url(url):
    """Read url."""
    req = requests.get(url)
    if not req.ok:
        _give_warning(req.content)
    return req.content


def _give_warning(req_content):
    root = ET.fromstring(req_content)
    exceptions = '\n'.join([" - " + ex_.text for ex_ in root.findall(EXCEPTION_TEXT)])
    exception_text = "\n\nFMI servers responded with the following errors:\n\n%s\n" % exceptions
    warnings.warn(exception_text)


def download_to_file(url, fname):
    """Download file from *ulr* to *fname*."""
    outpath, http_msg = urlretrieve(url, filename=fname)
    return outpath, http_msg



# -*- coding: utf-8 -*-
#
# Copyright (c) Panu Lahtinen
#
# Author(s):
#
#   Panu Lahtinen <pnuu+git@iki.fi>
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

import datetime as dt
import defusedxml.ElementTree as ET

from fmiopendata.utils import read_url

WMS_BASE = "https://openwms.fmi.fi/geoserver/wms?request=GetCapabilities"
WMS_LAYERS = './/{http://www.opengis.net/wms}Layer'
TIME_FORMAT = "%Y-%m-%dT%H:%M:%S.%fZ"


def get_wms_cababilities():
    """Get cababilities from WMS."""
    return read_url(WMS_BASE)


class WMSLayer(object):
    """Class for holding WMS layer information."""

    def __init__(self, layer):
        """Initialize layer."""
        self._layer = layer
        self.name = None
        self.title = None
        self.abstract = None
        self.crs = []
        self.bbox = []
        self.times = []
        self.elevations = None
        self.time_step_str = None
        self._parse_layer(layer)

    def __repr__(self):
        """Print WMS layer info."""
        if self.time_step_str is None:
            return self.name + " - no timesteps"
        if self.elevations is None:
            return self.name + " - " + self.time_step_str
        return self.name + " - " + self.time_step_str + ", elavations: " + ', '.join(self.elevations)

    def _get_times(self, txt):
        """Get timestamps."""
        start_time, end_time, step = txt.split('/')
        start_time = dt.datetime.strptime(start_time, TIME_FORMAT)
        end_time = dt.datetime.strptime(end_time, TIME_FORMAT)
        step = step.strip('PT')
        if step[-1] == 'H':
            tstep = dt.timedelta(hours=int(step[:-1]))
            self.time_step_str = step[:-1] + " hour time step"
        elif step[-1] == 'M':
            tstep = dt.timedelta(minutes=int(step[:-1]))
            self.time_step_str = step[:-1] + " minute time step"
        elif step[-1] == 'S':
            tstep = dt.timedelta(seconds=int(step[:-1]))
            self.time_step_str = step[:-1] + " second time step"
        else:
            return

        time_stamp = start_time
        while time_stamp <= end_time:
            self.times.append(time_stamp)
            time_stamp += tstep

    def _parse_layer(self, layer):
        for itm2 in list(layer):
            if "Name" in itm2.tag:
                self.name = itm2.text
            elif "Title" in itm2.tag:
                self.title = itm2.text
            elif "Abstract" in itm2.tag:
                self.abstract = itm2.text
            elif "CRS" in itm2.tag and "EPSG" in itm2.text:
                self.crs.append(itm2.text)
            elif "}BoundingBox" in itm2.tag and ("EPSG" in itm2.attrib['CRS'] or "CRS" in itm2.attrib['CRS']):
                self.bbox.append(itm2.attrib)
            elif "Dimension" in itm2.tag and itm2.attrib["name"] == "time":
                self._get_times(itm2.text)
            elif "Dimension" in itm2.tag and itm2.attrib["name"] == "elevation":
                self.elevations = itm2.text.split(',')


def get_wms_layers():
    """Get WMS layer info."""
    xml = read_url(WMS_BASE)
    root = ET.fromstring(xml)

    layers = root.findall(WMS_LAYERS)

    res = dict()
    for itm in layers:
        layer = WMSLayer(itm)
        if layer.name is not None:
            res[layer.name] = layer

    return res



# -*- coding: utf-8 -*-
#
# Copyright (c) Panu Lahtinen
#
# Author(s):
#
#   Panu Lahtinen <pnuu+git@iki.fi>
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

import datetime as dt
import defusedxml.ElementTree as ET
import os
import tempfile

import numpy as np
import eccodes

from fmiopendata import wfs
from fmiopendata.utils import read_url, download_to_file

TIME_FORMAT = "%Y-%m-%dT%H:%M:%SZ"


class Grid(object):
    """Class for holding grid data."""

    def __init__(self):
        """Initialize the class."""
        self.init_time = None
        self.start_time = None
        self.end_time = None
        self.data = dict()
        self.latitudes = None
        self.longitudes = None
        self.url = None
        self._fname = None

    def download(self, fname=None):
        """Read the data."""
        if self._fname is not None:
            return
        if fname is None:
            temp = tempfile.NamedTemporaryFile(delete=False)
            fname = temp.name
        outpath, http_msg = download_to_file(self.url, fname)
        self._fname = fname

    def parse(self, delete=False):
        """Parse the data."""
        if "grib" in self.url:
            parser = self._parse_grib
        else:
            format = [itm for itm in self.url.split("&") if "format" in itm.lower()][0]
            raise NotImplementedError("No parser for %s" % format)

        self.download()
        if self.data:
            return
        parser()
        if delete:
            self.delete_file()

    def _parse_grib(self):
        """Parser for GRIB data."""
        with eccodes.reader.FileReader(self._fname) as grib:
            for msg in grib:
                valid_date = msg["validityDate"]
                year = valid_date // 10000
                month = valid_date % 10000 // 100
                day = valid_date % 10000 % 100
                valid_time = msg["validityTime"]
                hour = valid_time // 100
                minute = valid_time % 100
                datime = dt.datetime(year, month, day, hour, minute)
                if self.latitudes is None:
                    self.latitudes = np.reshape(msg["latitudes"], (msg["Nj"], msg["Ni"]))
                    self.longitudes = np.reshape(msg["longitudes"], (msg["Nj"], msg["Ni"]))
                if datime not in self.data:
                    self.data[datime] = dict()
                if msg["level"] not in self.data[datime]:
                    self.data[datime][msg["level"]] = dict()
                level = self.data[datime][msg["level"]]
                level[msg["name"]] = dict()
                data = np.reshape(msg["values"], (msg["Nj"], msg["Ni"]))
                data[data == msg["missingValue"]] = np.nan
                level[msg["name"]]["data"] = data
                level[msg["name"]]["units"] = msg["units"]

    def delete_file(self):
        """Delete the downloaded file."""
        if os.path.isfile(self._fname):
            os.remove(self._fname)


class ParseGrids(object):
    """Class for parsing grid data."""

    def __init__(self, xml):
        """Initialize class."""
        self._xml = ET.fromstring(xml)
        self.data = dict()
        self._parse()

    def _parse(self):
        """Parse grid data."""
        for member in self._xml.findall(wfs.WFS_MEMBER):
            grid = Grid()
            grid.init_time = dt.datetime.strptime(member.findtext(wfs.GML_TIME_POSITION), TIME_FORMAT)
            grid.start_time = dt.datetime.strptime(member.findtext(wfs.GML_BEGIN_POSITION), TIME_FORMAT)
            grid.end_time = dt.datetime.strptime(member.findtext(wfs.GML_END_POSITION), TIME_FORMAT)
            grid.url = member.findtext(wfs.GML_FILE_REFERENCE)
            self.data[grid.init_time] = grid


def download_and_parse(query_id, args=None):
    """Download and parse the given stored query."""
    url = wfs.STORED_QUERY_URL + query_id
    if args:
        url = url + "&" + "&".join(args)
    xml = read_url(url)
    return ParseGrids(xml)



#!/usr/bin/env python
# -*- coding: utf-8 -*-
#
# Copyright (c) Panu Lahtinen
#
# Author(s):
#
#   Panu Lahtinen <pnuu+git@iki.fi>
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

"""Setup for trollflow2."""

from setuptools import setup
from fmiopendata import __version__

install_requires = ['numpy', 'requests', 'defusedxml']
extras_require = {'radar': ['rasterio'],
                  'grid': ['eccodes'],
                  }
all_extras = []
for extra_deps in extras_require.values():
    all_extras.extend(extra_deps)
extras_require['all'] = list(set(all_extras))

setup(name="fmiopendata",
      version=__version__,
      description='Python library for accessing FMI open data',
      long_description='Python library for accessing FMI open data',
      author='Panu Lahtinen',
      author_email='pnuu+git@iki.fi',
      classifiers=["Development Status :: 4 - Beta",
                   "Intended Audience :: Science/Research",
                   "License :: OSI Approved :: GNU General Public License v3 " +
                   "or later (GPLv3+)",
                   "Operating System :: OS Independent",
                   "Programming Language :: Python",
                   "Topic :: Scientific/Engineering",
                   ],
      url="https://github.com/pnuu/pyfmiopendata",
      packages=['fmiopendata', ],
      scripts=["bin/wms_html.py",
               "bin/wfs_html.py",
               ],
      data_files=[],
      install_requires=install_requires,
      extras_require=extras_require,
      tests_require=['rasterio', 'eccodes'],
      python_requires='>=3.8',
      )
